/** \file   dualization.c
 *  \author Takahisa Toda
 *  \see T.Toda, Dualization of Boolean functions Using Ternary Decision Diagrams, in Proc. of 13th International Symposium on Artificial Intelligence and Mathematics (ISAIM2014), Florida, USA(2014).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "my_def.h"
#include "my_random.h"
#include "my_opcache.h"
#include "my_io_dimacs.h"
#include "tdd_io_dimacs.h"
#include "tdd_trans.h"
#include "tdd_min.h"

#define PROGNAME "DBTDD"//!< the name of this program
#define PROGVERS "1.0.0"      //!< the current version

#ifndef NGSL
my_rng *g_rng; //!< random number generator for GSL
#endif /*NGSL*/

#define PRINT_USAGE(p) do{fprintf(stderr, "Usage: %s input-file [output-file]\n", (p));	\
    fprintf(stderr, "%s\tVersion %s\n", PROGNAME, PROGVERS);		\
  } while(0)


int main(int argc, char *argv[])
{  
  int res = MY_SUCCESS;
  char *infile = NULL;
  char *outfile = NULL;
  
  /*** RECEIVE INPUTS ***/  
  for(int i = 1; i < argc; i++) {
    if(argv[i][0] == '-') {
      switch (argv[i][1]){
      case 'v': case '?': case 'h': default:
        PRINT_USAGE(argv[0]); return  MY_SUCCESS;
      }
    } else {
      if(infile == NULL) {infile = argv[i];}
      else if(outfile == NULL) {outfile = argv[i];}
      else {PRINT_USAGE(argv[0]); return  MY_SUCCESS;}
    }
  }
  if(infile == NULL) {PRINT_USAGE(argv[0]); return  MY_SUCCESS;}
#ifdef LOG_INFO
  printf("%s\t\tversion %s\n", PROGNAME, PROGVERS);
  printf("package\t\t%s\n", ZTDD_PACKAGE);
  printf("package\t\t%s\n", DTDD_PACKAGE);
  printf("input\t\t%s\n", infile);
  if(outfile != NULL) printf("output\t\t%s\n", outfile);
#endif /*LOG_INFO*/

  my_initrand(); // initialize random number generator
  FILE *in = fopen(infile, "r");
  ENSURE_TRUE_MSG(in != NULL, "file open failed\n");
  
  /*** INITIALIZE TDD ENVIRONMENT ***/
  uintmax_t nrow, total_nent;
  my_val maxval;
  res = getfileinfo_dimacs(&nrow, &maxval, &total_nent, in);
  ENSURE_SUCCESS(res);
  // adjust OPCACHE_DEFAULT_SIZE if necessary.
  ztdd_init(maxval, OPCACHE_DEFAULT_SIZE);
  dtdd_init(maxval, OPCACHE_DEFAULT_SIZE);
  
  /*** IMPORT Z-TDD ***/
  clock_t startingtime = my_clock();
  my_ztddp f = ztdd_import_dimacs(nrow, total_nent, in);
  ENSURE_TRUE_MSG(f != ZTDD_NULL, "Z-TDD import failed\n");
  print_elapsed_time(startingtime, "IMPORT");
#ifdef LOG_INFO
  printf("|ztdd|\t\t%ju\n", ztdd_size(f));
  printf("#sets\t\t%ju\n", ztdd_card(f));
  printf("total\t\t%ju\n\n", ztdd_totalsize(f));
  fflush(stdout);
#endif /*LOG_INFO*/

  /*** DUALIZATION ***/
  startingtime = my_clock();
  my_dtddp g = trans_z2d(f);
  ENSURE_TRUE(g != DTDD_NULL);
  print_elapsed_time(startingtime, "TRANS");
#ifdef LOG_INFO
  printf("|dtdd|\t\t%ju\n", dtdd_size(g));
  //printf("#sets\t\t%ju\n", dtdd_card(maxval, g));
  //printf("total\t\t%ju\n\n", dtdd_totalsize(maxval, g));
  fflush(stdout);
#endif /*LOG_INFO*/

  startingtime = my_clock();
  my_ztddp r = min_d2z(g);
  ENSURE_TRUE(r != ZTDD_NULL);
  print_elapsed_time(startingtime, "MIN");
#ifdef LOG_INFO
  printf("|ztdd|\t\t%ju\n", ztdd_size(r));
  printf("#sets\t\t%ju\n", ztdd_card(r));
  printf("total\t\t%ju\n\n", ztdd_totalsize(r));
  fflush(stdout);
#endif /*LOG_INFO*/

  /*** EXPORT Z-TDD ***/
  if(outfile != NULL) {
    FILE *out = fopen(outfile, "w");
    ENSURE_TRUE_MSG(out != NULL, "file open failed\n");
    startingtime = my_clock();
    ztdd_export_dimacs(maxval, r, out, outfile);
    //ztdd_export_dot(maxval, r, out);
    print_elapsed_time(startingtime, "EXPORT");
    fclose(out);
  }

  ztdd_quit(); dtdd_quit();
  my_freerand();
  fclose(in);

  return MY_SUCCESS;
}
