/** \file   dd_wrapper.h
 *  \brief  Functions which uses DD package.
 *  \author Takahisa Toda
 */
#ifndef DD_WRAPPER_H
#define DD_WRAPPER_H

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#include "my_def.h"

#ifdef SELECT_MYTDD
/*---------------------------------------------------------------------------*/
/* MY TDD PACKAGE                                                            */
/*---------------------------------------------------------------------------*/
#include "my_tdd.h"

#define ZTDD_PACKAGE "My Z-TDD Package"
#define DTDD_PACKAGE "My D-TDD Package"
#define ZTDD_NULL  MYTDD_NULL
#define DTDD_NULL  MYTDD_NULL
#define ZTDD_VAR_MAXINDEX  MYTDD_VARMAX //!< maximum value that a Z-TDD library can handle.
#define DTDD_VAR_MAXINDEX  MYTDD_VARMAX //!< maximum value that a D-TDD library can handle.

typedef struct mytdd_node *my_ztddp; //!< pointer to a Z-TDD node
typedef struct mytdd_node *my_dtddp; //!< pointer to a D-TDD node

typedef struct mytdd_nodearray my_ztddarray;
typedef struct mytdd_nodearray my_dtddarray;

/* Operations for Z-TDDs */
static inline int ztdd_init(my_val maxval, uintmax_t opcache_size)
{
  ENSURE_TRUE_MSG(0 < maxval && maxval <= ZTDD_VAR_MAXINDEX && maxval < INFTY, "The number of variable invalid");
  int res = mytdd_unique_table_create(opcache_size);
#ifdef LOG_INFO
  printf("ZTDD_VAR_MAXINDEX:\t%ju\n", (uintmax_t)ZTDD_VAR_MAXINDEX);
#endif /*LOG_INFO*/
  return res;
}

static inline int ztdd_quit(void)
{
  return mytdd_unique_table_destroy();
}

static inline uintmax_t ztdd_size(my_ztddp f) // the number of Z-TDD internal nodes
{
  return mytdd_size(f); 
}

static inline uintmax_t ztdd_totalsize(my_ztddp f)
{
  return mytdd_totalsize(f);
}

static inline uintmax_t ztdd_card(my_ztddp f) // the number of paths from the root to 1 terminal node
{
  return mytdd_card(f);
}

static inline uintmax_t ztdd_uniquetable_maxdepth(void)
{
  return mytdd_unique_table_maxdepth();
}

static inline my_ztddp ztdd_bot(void)
{
  return mytdd_bot;
}

static inline my_ztddp ztdd_top(void)
{
  return mytdd_top;
}

static inline bool ztdd_isconst(my_ztddp f)
{
  return (f==ztdd_bot() || f==ztdd_top());
}

static inline my_val ztdd_label(my_ztddp f)
{
  assert(ZTDD_NULL != f);
  return f->k;
}

static inline my_ztddp ztdd_POS(my_ztddp f)
{
  assert(ZTDD_NULL != f);
  assert(!ztdd_isconst(f));
  return f->pos;
}

static inline my_ztddp ztdd_NEG(my_ztddp f)
{
  assert(ZTDD_NULL != f);
  assert(!ztdd_isconst(f));
  return f->neg;
}


static inline my_ztddp ztdd_ZERO(my_ztddp f)
{
  assert(ZTDD_NULL != f);
  assert(!ztdd_isconst(f));
  return f->zero;
}

static inline uintptr_t ztdd_AUX(my_ztddp f)
{
  assert(ZTDD_NULL != f);
  return f->aux;
}

static inline void ztdd_setAUX(uintptr_t k, my_ztddp f)
{
  assert(ZTDD_NULL != f);
  mytdd_setaux(k, f);
}


static inline my_ztddp ztdd_diff(my_ztddp f, my_ztddp g)
{
  return myztdd_diff(f, g);
}

static inline my_ztddp ztdd_getnode(my_val v, my_ztddp zero, my_ztddp neg, my_ztddp pos)
{
  return myztdd_getnode(v, zero, neg, pos);
}

static inline my_ztddarray *ztdd_create_array_with_nodes_levelwise(my_val maxval, my_ztddp f)
{
  return mytdd_create_array_with_nodes_levelwise(maxval, f);
}

static inline void ztdd_destroy_array_with_nodes_levelwise(my_ztddarray *na)
{
  mytdd_destroy_array_with_nodes_levelwise(na);
}

static inline int ztdd_export_dot(my_val maxval, my_ztddp f, FILE *out)
{
  return tdd_export_dot(maxval, f, out);
}

/* Operations for D-TDDs */
static inline int dtdd_init(my_val maxval, uintmax_t opcache_size)
{
  ENSURE_TRUE_MSG(0 < maxval && maxval <= DTDD_VAR_MAXINDEX && maxval < INFTY, "The number of variable invalid");
  int res = mytdd_unique_table_create(opcache_size);
#ifdef LOG_INFO
  printf("DTDD_VAR_MAXINDEX:\t%ju\n", (uintmax_t)DTDD_VAR_MAXINDEX);
#endif /*LOG_INFO*/
  return res;
}

static inline int dtdd_quit(void)
{
  return mytdd_unique_table_destroy();
}

static inline uintmax_t dtdd_size(my_dtddp f) // the number of D-TDD internal nodes
{
  return mytdd_size(f);
}

static inline uintmax_t dtdd_card(my_val maxval, my_dtddp f) // the number of paths from the root to 1 terminal node
{
  struct mytdd_node *g = mytdd_d2q(maxval, f);
  return mytdd_card(g);
}

static inline uintmax_t dtdd_totalsize(my_val maxval, my_dtddp f)
{
  struct mytdd_node *g = mytdd_d2q(maxval, f);
  return mytdd_totalsize(g);
}

static inline uintmax_t dtdd_uniquetable_maxdepth(void)
{
  return  mytdd_unique_table_maxdepth();
}

static inline my_dtddp dtdd_bot(void)
{
  return mytdd_bot;
}

static inline my_dtddp dtdd_top(void)
{
  return mytdd_top;
}

static inline bool dtdd_isconst(my_dtddp f)
{
  return (f==dtdd_bot() || f==dtdd_top());
}

static inline my_val dtdd_label(my_dtddp f)
{
  assert(DTDD_NULL != f);
  return f->k;
}

static inline my_dtddp dtdd_POS(my_dtddp f)
{
  assert(DTDD_NULL != f);
  assert(!dtdd_isconst(f));
  return f->pos;
}

static inline my_dtddp dtdd_NEG(my_dtddp f)
{
  assert(DTDD_NULL != f);
  assert(!dtdd_isconst(f));
  return f->neg;
}


static inline my_dtddp dtdd_ZERO(my_dtddp f)
{
  assert(DTDD_NULL != f);
  assert(!dtdd_isconst(f));
  return f->zero;
}

static inline uintptr_t dtdd_AUX(my_dtddp f)
{
  assert(DTDD_NULL != f);
  return f->aux;
}

static inline void dtdd_setAUX(uintptr_t k, my_dtddp f)
{
  assert(DTDD_NULL != f);
  mytdd_setaux(k, f);
}


static inline my_dtddp dtdd_intersect(my_dtddp f, my_dtddp g)
{
  return mydtdd_intersect(f, g);
}

static inline my_dtddp dtdd_getnode(my_val v, my_dtddp zero, my_dtddp neg, my_dtddp pos)
{
  return mydtdd_getnode(v, zero, neg, pos);
}

static inline my_dtddarray *dtdd_create_array_with_nodes_levelwise(my_val maxval, my_dtddp f)
{
  return mytdd_create_array_with_nodes_levelwise(maxval, f);
}

static inline void dtdd_destroy_array_with_nodes_levelwise(my_dtddarray *na)
{
  mytdd_destroy_array_with_nodes_levelwise(na);
}

static inline int dtdd_export_dot(my_val maxval, my_dtddp f, FILE *out)
{
  return tdd_export_dot(maxval, f, out);
}

#else
#error "Specify TDD package"
#endif

#endif /*DD_WRAPPER_H*/
