/** \file   my_hash.c
 *  \brief  Implementation of a simple hash table
 *  \author Takahisa Toda
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "my_def.h"
#include "my_hash.h"


/** \brief  Create a hash table.
 *  \param  nbuckets number of buckets
 *  \return pointer to a hash table if successful; otherwise, NULL.
 *  \see    ht_destroy, ht_enlarge
 */
my_hash *ht_create(uintmax_t nbuckets)
{
  ENSURE_TRUE(0 < nbuckets);
  my_hash *h = (my_hash*) malloc(sizeof(my_hash));
  ENSURE_TRUE_MSG(h != NULL,"memory allocation failed");

  h->nbucket = nbuckets;
  h->bucket = (struct hash_entry**) malloc(sizeof(struct hash_entry*) * nbuckets);
  ENSURE_TRUE_MSG(h->bucket != NULL, "memory allocation failed");
  for (uintmax_t i = 0; i < nbuckets; i++) h->bucket[i] = NULL;

  return h;
}


/** \brief Free memory area used by a hash table.
 *  \param h pointer to a hash table
 *  \see   ht_create, ht_enlarge
 */
void ht_destroy(my_hash *h)
{
  assert(h != NULL);
  const uintmax_t nbucket = h->nbucket;
  for(uintmax_t i = 0; i < nbucket; i++) {
      for(struct hash_entry *c = h->bucket[i]; c != NULL; ) {
	struct hash_entry *t = c->next;
	free(c);
	c = t;
      }
  }
  free(h->bucket); h->bucket = NULL;
  free(h);         h = NULL;
}

/** \brief  Calculate a hash value.
 *  \param  k  key
 *  \param  size the size of a hash table, i.e. the number of buckets
 *  \return the calculated hash value
 */
static inline uintmax_t my_calhashval(uintptr_t k, uintmax_t size)
{
  assert(size > 0);
  return (uintmax_t)(k*MY_PRIME(0))%size;
}


/** \brief  Register a new entry to the tail of a linked list.
 *  \param  k  key
 *  \param  d  data associated with key
 *  \param  h  pointer to a hash table
 *  \see    ht_search
 */
void ht_add(uintptr_t k, uintptr_t d, my_hash *h)
{
  assert(h != NULL);
  struct hash_entry *c = (struct hash_entry*)malloc(sizeof(struct hash_entry));
  ENSURE_TRUE_MSG(c != NULL, "memory allocation failed");
  c->key = k;
  c->data = d;

  const uintmax_t val = my_calhashval(k, h->nbucket);
  c->next = h->bucket[val];
  h->bucket[val] = c;
}


/** \brief  Find a hash entry. If found, copy the associated data to pd.
 *  \param  k    key
 *  \param  pd   pointer to a variable, in which data is copied.
 *  \param  h    pointer to a hash table
 *  \return MY_SUCCESS if found; otherwise, MY_FAILURE.
 *  \see    ht_add
 *  \note
 *  - data is not copied if pd == NULL.
 */
int ht_search(uintptr_t k, uintptr_t *pd, const my_hash *h)
{
  assert(h != NULL);
  const uintmax_t val = my_calhashval(k, h->nbucket);
  for(struct hash_entry *c = h->bucket[val]; c != NULL; c = c->next) {
    if(c->key == k) {
      if(pd != NULL) *pd = c->data;
      return MY_SUCCESS;
    }
  }
  return MY_FAILURE;
}


/* \brief Compute the maximum length of a linked list.
 *  \param h pointer to a hash table
 *  \return the computed maximum length
 *  \see ht_create, ht_destroy
 */
uintmax_t ht_maxdepth(const my_hash *h)
{
  assert(h != NULL);
  uintmax_t depth = 0;
  const uintmax_t nbucket = h->nbucket;
  for(uintmax_t i = 0; i < nbucket; i++) {
      uintmax_t j = 0;
      for(struct hash_entry *t = h->bucket[i]; t != NULL; t=t->next) {j++;}
      if(depth < j) {depth = j;}
    }
  return depth;
}

/* \brief Enlarge hash table size
 * \param ch pointer to a hash table
 * \see ht_create, ht_destroy
 */
void ht_enlarge(my_hash *h)
{
  const uintmax_t oldsiz = h->nbucket;
  ENSURE_TRUE_MSG(oldsiz < UINTMAX_MAX / 4, "hash table size overflow");
  const uintmax_t newsiz = oldsiz << 2;
#ifdef LOG_INFO
  printf("hash table resizing:%ju -> %ju\n", oldsiz, newsiz);
#endif /*LOG_INFO*/
  struct hash_entry **oldbucket = h->bucket;
  struct hash_entry **newbucket = (struct hash_entry**) malloc(sizeof(struct hash_entry*) * newsiz);
  ENSURE_TRUE_MSG(newbucket != NULL, "memory allocation failed");
  for(uintmax_t i = 0; i < newsiz; i++) newbucket[i] = NULL;
  for(uintmax_t i = 0; i < oldsiz; i++) {
    for(struct hash_entry *c = oldbucket[i]; c != NULL;) {
      struct hash_entry *nextnode = c->next;
      const uintmax_t val = my_calhashval(c->key, newsiz);
      c->next = newbucket[val];
      newbucket[val] = c;      
      c = nextnode;
    }
  }
  free(oldbucket); oldbucket = NULL;
  h->nbucket = newsiz;
  h->bucket = newbucket;
}
