/** \file   my_io_dimacs.c
 *  \brief  Implementation of IO functions for data files in DIMACS cnf format
 *  \author Takahisa Toda
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>

#include "my_def.h"
#include "my_io_dimacs.h"

static inline int my_comp_int_wrt_abs(const void *a, const void *b)
{
  const my_val c = *((const my_val*)a);
  const my_val d = *((const my_val*)b);
  
  if (abs(c) > abs(d))
    return 1;
  else if (abs(c) < abs(d))
    return -1;
  else
    return 0;
}


/**  \brief Obtain information of a data file in the DIMACS format, and check consistency of file format.
 *  \param  pnrows       pointer to the number of rows, including empty rows.
 *  \param  pmaxval      pointer to the maximum value of an entry
 *  \param  ptotal_nent  pointer to the total number of entries in a datafile
 *  \param  in           pointer to a dafa file
 *  \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 */
int getfileinfo_dimacs(uintmax_t *pnrows, my_val *pmaxval, uintmax_t *ptotal_nent, FILE *in)
{
  assert(pnrows != NULL); assert(pmaxval != NULL); assert(ptotal_nent != NULL); assert(in != NULL);
  fseek(in, 0L, SEEK_SET);
  *pnrows = *pmaxval = *ptotal_nent = 0;
  char buf[BUFSIZ];
  char c;  char *p = buf; uintmax_t nc = 0;
  bool pline_appeared = false; bool EOF_appeared = false;

  while (!EOF_appeared && (c = fgetc(in)) != EOF) {
    switch (c) {
    case DIMACS_COMMENT: // skip comment line
      do {;} while((c = fgetc(in)) != '\n' && c != EOF);
      if(c == EOF) EOF_appeared = true;
      break;
    case DIMACS_PROBLEM: // read problem line
      ENSURE_TRUE_MSG(!pline_appeared, "problem line already appeared.");
      pline_appeared = true;
      do {assert(p < buf + BUFSIZ); *p = c; p++;} while((c = fgetc(in)) != '\n' && c != EOF) ;
      if(c == EOF) EOF_appeared = true;
      char tc; char ts[BUFSIZ]; 
      sscanf(buf, "%c %s %jd %ju", &tc, ts, pmaxval, pnrows);
      p = buf;
      break;
    default:
      ENSURE_TRUE_MSG(pline_appeared, "problem line does not appeared yet.");
      while(c == DIMACS_SEPARATOR_WS || c == DIMACS_SEPARATOR_TAB || c == '\n') {c = fgetc(in);}
      if(c == EOF) {EOF_appeared = true; break;}
      while(c != DIMACS_SEPARATOR_WS && c!= DIMACS_SEPARATOR_TAB && c != '\n' && c != EOF) {
	assert(p < buf + BUFSIZ);
	*p = c; p++;
	c = fgetc(in);
      }
      assert(p < buf + BUFSIZ);
      *p = '\0';
      my_val entry = atoi(buf);
      if(entry == 0) {nc++; assert(nc <= *pnrows);}
      else (*ptotal_nent)++;
      p = buf;
      if(c == EOF) EOF_appeared = true;
      break;
    }
  }
  ENSURE_TRUE_WARN(*pnrows == nc, "the number of clauses in the problem line is not correct.");
  *pnrows = nc;
#ifdef LOG_INFO
  printf("maxval: %jd, nrows: %ju\n", *pmaxval, *pnrows);
#endif /*LOG_INFO*/
  return MY_SUCCESS;
}


/**  \brief Convert a data in the DIMACS format into an internal representation.
 *  \param  nrows       the number of rows
 *  \param  total_nent  the total number of entries in a data file
 *  \param  in          pointer to a dafa file
 *  \note
 *  - the j-th entry in the i-th row corresponds to sf->elem[i-1][j-1], thus 1 <= i <= nrows.
 *  - If the i-th row contains k entries, then 1<= j <= k.
 *  - Furthermore, sf->elem[i-1][k] = INFTY.
 *  - the returned data structure must be destroyed when it becomes unnecessary.
 */
struct setfam *in2setfam_dimacs(uintmax_t nrows, uintmax_t total_nent, FILE *in)
{
  assert(in != NULL);
  fseek(in, 0L, SEEK_SET);

  const uintmax_t memsize = total_nent + nrows;
  struct setfam *sf = (struct setfam*)malloc(sizeof(struct setfam));
  ENSURE_TRUE_MSG(sf != NULL, "memory allocation failed");
  sf->elem = (my_val**)malloc(sizeof(my_val*)*nrows);
  ENSURE_TRUE_MSG(sf->elem != NULL, "memory allocation failed");
  sf->mem = (my_val*)malloc(sizeof(my_val)*memsize);
  ENSURE_TRUE_MSG(sf->mem != NULL, "memory allocation failed");
  sf->nsets = nrows;
  sf->memsize = memsize;

  my_val **ee = sf->elem; my_val *mm = sf->mem;
  my_val *first_pos = mm; // the first position of each row
  char buf[BUFSIZ]; char *p = buf; char c;
  bool pline_appeared = false; bool EOF_appeared = false;

  while (!EOF_appeared && (c = fgetc(in)) != EOF) {
    switch (c) {
    case DIMACS_COMMENT: // skip comment line
      do {;} while((c = fgetc(in)) != '\n' && c != EOF);
      if(c == EOF) EOF_appeared = true;
      break;
    case DIMACS_PROBLEM: // skip problem line
      ENSURE_TRUE_MSG(!pline_appeared, "problem line already appeared.");
      pline_appeared = true;
      do {;} while((c = fgetc(in)) != '\n' && c != EOF);
      if(c == EOF) EOF_appeared = true;
      break;
    default:
      ENSURE_TRUE_MSG(pline_appeared, "problem line does not appeared yet.");
      while(c == DIMACS_SEPARATOR_WS || c == DIMACS_SEPARATOR_TAB || c == '\n') {c = fgetc(in);}
      if(c == EOF) {EOF_appeared = true; break;}
      while(c != DIMACS_SEPARATOR_WS && c!= DIMACS_SEPARATOR_TAB && c != '\n' && c != EOF) {
	assert(p < buf + BUFSIZ);
	*p = c; p++;
	c = fgetc(in);
      }
      assert(p < buf + BUFSIZ);
      *p = '\0';
      my_val entry = atoi(buf);
      if (entry == 0) {
	assert(mm < sf->mem + memsize);
	*mm = INFTY; mm++;
	assert(ee < sf->elem + nrows);
	*ee = first_pos; ee++; // register the current row after all entries were read.
	first_pos = mm; // update first_pos to be the first position of the next row.
      } else {
	assert(mm < sf->mem + memsize);
	*mm = entry; mm++;
      }
      p = buf;
      if(c == EOF) EOF_appeared = true;
      break;
    }
  }
  assert(mm == sf->mem + memsize);
  assert(ee == sf->elem + nrows);
  return sf;
}


/* \brief Free all memory areas used.
 * \param sf pointer to setfam
 */
void destroy_setfam(struct setfam *sf)
{
  assert(sf != NULL);
  free(sf->elem); free(sf->mem); free(sf);
}



void format_eachrow(my_val *x[], uintmax_t nrows)
{
  for(uintmax_t i = 0; i < nrows; i++) {
    uintmax_t depth = 0; while(x[i][depth] != INFTY) {depth++;}
    qsort(x[i], depth, sizeof(my_val), my_comp_int_wrt_abs);
    my_val prev = 0;
    for(uintmax_t j = 0; j < depth; j++) {
      ENSURE_TRUE_MSG(abs(prev) != abs(x[i][j]) || prev != x[i][j], "duplicated entries");
      prev = x[i][j];
    }
  }
}
