/** \file   my_opcache.c
 *  \brief  Implementation of a simple cache for binary operations
 *  \author Takahisa Toda
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "my_def.h"
#include "my_opcache.h"

enum CACHE_POSKEY {
  OPCACHE_POS_OPID = 0,
  OPCACHE_POS_OPERAND1,
  OPCACHE_POS_OPERAND2,
  OPCACHE_POS_RESULT,
  OPCACHE_ENTRY_SIZE, /* if you want to add a new field in a cache entry, insert it before OPCACHE_ENTRY_SIZE.*/
};


/** \brief  Create a cache.
 *  \param  nent the number of cache entries
 *  \return pointer to a cache if successful; otherwise, NULL.
 *  \see    opcache_destroy, opcache_enlarge
 */
my_opcache *opcache_create(uintmax_t nent)
{
  ENSURE_TRUE(0 < nent);
  my_opcache *ca = (my_opcache*) malloc(sizeof(my_opcache));
  ENSURE_TRUE_MSG(ca != NULL,"memory allocation failed");

  ca->nent = nent;
  const uintmax_t len = nent*OPCACHE_ENTRY_SIZE;
  ca->entry = (uintptr_t*) malloc(sizeof(uintptr_t) * len);
  ENSURE_TRUE_MSG(ca->entry != NULL, "memory allocation failed");
  for(uintmax_t i = 0; i < len; i++) ca->entry[i] = 0;

  return ca;
}


/** \brief Free memory area used by a cache.
 *  \param ca pointer to a cache
 *  \see   opcache_create, opcache_enlarge
 */
void opcache_destroy(my_opcache *ca)
{
  ENSURE_TRUE(ca != NULL);
  free(ca->entry); ca->entry = NULL;
  free(ca);        ca = NULL;
}

/** \brief  Calculate a hash value.
 *  \param  id   operation id (>0)
 *  \param  k1   operand
 *  \param  k2   operand
 *  \param  num  the total number of cache entries
 *  \return the calculated hash value, which must be less than the number of cache entries.
 *  \note   The position of a cache entry can be computed by multipling a hash value by OPCACHE_ENTRY_SIZE.
 */
static inline uintmax_t my_calhashval(int id, uintptr_t k1, uintptr_t k2, uintmax_t num)
{
  assert(num > 0);
  return (id*MY_PRIME(0) + (uintmax_t)k1*MY_PRIME(1) + (uintmax_t)k2*MY_PRIME(2))%num;
}


/** \brief  Register a new entry.
 *  \param  id   operation id (> 0)
 *  \param  k1   operand
 *  \param  k2   operand
 *  \param  r    result
 *  \see    opcache_search
 *  \note   Overwrite if the position is already filled.
 */
void opcache_add(int id, uintptr_t k1, uintptr_t k2, uintptr_t r, my_opcache *ca)
{
  assert(id > 0);
  const uintmax_t basepos = OPCACHE_ENTRY_SIZE * my_calhashval(id, k1, k2, ca->nent);
  ca->entry[basepos + OPCACHE_POS_OPID]     =  id;
  ca->entry[basepos + OPCACHE_POS_OPERAND1] =  k1;
  ca->entry[basepos + OPCACHE_POS_OPERAND2] =  k2;
  ca->entry[basepos + OPCACHE_POS_RESULT]   =  r;
}


/** \brief  Find a cache entry. If found, copy the operation result to pr.
 *  \param  id   operation id (> 0)
 *  \param  k1   operand
 *  \param  k2   operand
 *  \param  pr   pointer to the operation result
 *  \param  ca   pointer to a cache
 *  \return MY_SUCCESS if found; otherwise, MY_FAILURE.
 *  \see    opcache_add
 *  \note
 *  - result is not copied if pr == NULL.
 */
int opcache_search(int id, uintptr_t k1, uintptr_t k2, uintptr_t *pr, my_opcache *ca)
{
  assert(id > 0);
  const uintmax_t basepos = OPCACHE_ENTRY_SIZE * my_calhashval(id, k1, k2, ca->nent);
  if(ca->entry[basepos+OPCACHE_POS_OPID] == id 
     && ca->entry[basepos+OPCACHE_POS_OPERAND1] == k1 
     && ca->entry[basepos+OPCACHE_POS_OPERAND2] == k2) {
    if(pr != NULL) *pr = ca->entry[basepos+OPCACHE_POS_RESULT];
    return MY_SUCCESS;
  } else {return MY_FAILURE;}
}


/* \brief Enlarge cache size.
 * \param ca pointer to a cache
 * \see opcache_create, opcache_destroy
 */
void opcache_enlarge(my_opcache *ca)
{
  const uintmax_t oldsiz = ca->nent; 
  const uintmax_t newsiz = oldsiz << 2; const uintmax_t newlen = OPCACHE_ENTRY_SIZE*newsiz;
#ifdef LOG_INFO
  printf("opcache resizing:%ju -> %ju\n", oldsiz, newsiz);
#endif /*LOG_INFO*/
  uintptr_t *oldmem = ca->entry;
  uintptr_t *newmem = (uintptr_t*) malloc(sizeof(uintptr_t) * newlen);
  ENSURE_TRUE_MSG(newmem != NULL, "memory allocation failed");
  for(uintmax_t i = 0; i < newlen; i++) newmem[i] = 0;
  for(uintmax_t i = 0; i < oldsiz; i++) {
    const uintmax_t oldbasepos = OPCACHE_ENTRY_SIZE * i;
    const uintmax_t newbasepos = OPCACHE_ENTRY_SIZE * my_calhashval(oldmem[oldbasepos+OPCACHE_POS_OPID], 
								  oldmem[oldbasepos+OPCACHE_POS_OPERAND1], 
								  oldmem[oldbasepos+OPCACHE_POS_OPERAND2], 
								  newsiz);
    newmem[newbasepos+OPCACHE_POS_OPID]     = oldmem[oldbasepos+OPCACHE_POS_OPID];
    newmem[newbasepos+OPCACHE_POS_OPERAND1] = oldmem[oldbasepos+OPCACHE_POS_OPERAND1];
    newmem[newbasepos+OPCACHE_POS_OPERAND2] = oldmem[oldbasepos+OPCACHE_POS_OPERAND2];
    newmem[newbasepos+OPCACHE_POS_RESULT]   = oldmem[oldbasepos+OPCACHE_POS_RESULT];
  }
  free(oldmem); oldmem = NULL;
  ca->nent  = newsiz;
  ca->entry = newmem;
}


