/** \file   my_tdd.h
 *  \brief  TDD package
 *  \author Takahisa Toda
 */
#ifndef MY_TDD_H
#define MY_TDD_H

#include <stdlib.h>
#include <limits.h>
#include "my_def.h"

#define MYTDD_NULL                           NULL
#define MYTDD_VARMAX                         (INT32_MAX-1)
#define MYTDD_UNIQUE_TABLE_DEFAULT_SIZE      (UINTMAX_C(1) << 24)
#define MYTDD_UNIQUE_TABLE_DEPTH_THRESHOLD   (10)
#define MYTDD_MAX_EXCESSVAL_DEPTH_THRESHOLD  (20)

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/** \brief decidion diagram node*/
struct mytdd_node {
  my_val k;    //!< key
  bool visited; //!< true: terminal nodes (always), false: internal nodes (default)
  uintmax_t aux;     //!< an auxiliary field
  struct mytdd_node *zero; //!< zero arc
  struct mytdd_node *neg;  //!< negative arc
  struct mytdd_node *pos;  //!< positive arc
  struct mytdd_node *nx;   //!< internal use only
};

struct mytdd_nodearray {
  struct mytdd_node **a;  //!< array of TDD nodes
  uintmax_t *b;           //!< array that holds the size of each level: b[i] holds the number of nodes in level i, where 1 <= i <= nlevels.
  uintmax_t len;          //!< the length of a
  uintmax_t nlevels;      //!< the number of levels
};


extern struct mytdd_node *mytdd_top; //!< top terminal node
extern struct mytdd_node *mytdd_bot; //!< bottom terminal node

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

// common operations
extern int mytdd_unique_table_create(uintmax_t opcache_size);
extern int mytdd_unique_table_destroy(void);
extern uintmax_t mytdd_size(struct mytdd_node *f);
extern uintmax_t mytdd_unique_table_maxdepth(void);
extern struct mytdd_nodearray *mytdd_create_array_with_nodes_levelwise(my_val maxval, struct mytdd_node *f);
extern void mytdd_destroy_array_with_nodes_levelwise(struct mytdd_nodearray *na);
extern int tdd_export_dot(my_val maxval, struct mytdd_node *f, FILE *out);
extern uintmax_t mytdd_card(struct mytdd_node *f);
extern uintmax_t mytdd_totalsize(struct mytdd_node *f);
extern struct mytdd_node *mytdd_d2q(my_val maxval, struct mytdd_node *f);

static inline void mytdd_setaux(uintptr_t k, struct mytdd_node *f)
{
  f->aux = k;
}

// operations for D-TDDs
extern struct mytdd_node *mydtdd_getnode(my_val k, struct mytdd_node *zero, struct mytdd_node *neg, struct mytdd_node *pos);
extern struct mytdd_node *mydtdd_intersect(struct mytdd_node *f, struct mytdd_node *g);

// operations for Z-TDDs
extern struct mytdd_node *myztdd_getnode(my_val k, struct mytdd_node *zero, struct mytdd_node *neg, struct mytdd_node *pos);
extern struct mytdd_node *myztdd_diff(struct mytdd_node *f, struct mytdd_node *g);

// operations for Q-TDDs
extern struct mytdd_node *myqtdd_getnode(my_val k, struct mytdd_node *zero, struct mytdd_node *neg, struct mytdd_node *pos);

#ifdef MY_DEBUG
extern int mytdd_check_consistency(struct mytdd_node *f);
#endif /*MY_DEBUG*/

#endif /*MY_TDD_H*/
