/** \file   tdd_io_dimacs.c
 *  \brief  Implementation of IO functions for data files in DIMACS cnf format
 *  \see T.Toda, Dualization of Boolean functions Using Ternary Decision Diagrams, in Proc. of 13th International Symposium on Artificial Intelligence and Mathematics (ISAIM2014), Florida, USA(2014).
 */
#include <string.h>
#include <time.h>

#include "my_io_dimacs.h"
#include "tdd_io_dimacs.h"
#include "my_random.h"
#include "ssort2.h"

static inline uintmax_t my_max(uintmax_t a, uintmax_t b)
{
  return ((a)<=(b) ? (b) : (a));
}

static my_ztddp tpconstruct(my_val *x[], uintmax_t n, uintmax_t depth);

/** \brief  Build a Z-TDD from input data in DIMACS cnf format
 *  \param  nrows       the number of rows
 *  \param  total_nent  the total number of entries in a data file
 *  \param  in          pointer to a dafa file
 *  \return a root node
 */
my_ztddp ztdd_import_dimacs(uintmax_t nrows, uintmax_t total_nent, FILE *in)
{
  assert(in != NULL);
  struct setfam *sf = in2setfam_dimacs(nrows, total_nent, in);
  ENSURE_TRUE(sf != NULL);
  format_eachrow(sf->elem, nrows);
  ssort2main(sf->elem, nrows);
  my_ztddp f = tpconstruct(sf->elem, nrows, 0);
  ENSURE_TRUE(f != ZTDD_NULL);
  destroy_setfam(sf);
  return f;
}


/** \brief Find the first position with depth-th value not equal to v (by a binary search).
 * \param x      x[i] corresponds to the i-th row in a data file
 * \param left   the leftend position
 * \param right  the rightend position
 * \param v      value
 * \param depth  the position of rows to be examined.
 * \return the found position.
 * \note assume that (v ==) x[left][depth] <= x[left+1][depth] <= ... <= x[right][depth] (!= v)
 */
static inline uintmax_t my_bsearch(my_val *x[], uintmax_t left, uintmax_t right, const my_val v, const uintmax_t depth)
{
  assert(x != NULL);
  while(1) {
    assert(x[left][depth] == v); assert(x[right][depth] != v); assert(left < right);
    if(right - left <= 1) return right;
    uintmax_t mid = (left + right) / 2;
    if(x[mid][depth] == v) left = mid;
    else right = mid;
  }
}

/** \brief Build a Z-TDD from input data
 * \param x      x[i] corresponds to the i-th row in a data file
 * \param n      the number of rows
 * \param depth  the position of rows to be examined.
 * \return a root node
 */
static my_ztddp tpconstruct(my_val *x[], uintmax_t n, uintmax_t depth)
{ 
  assert(x != NULL);
  uintmax_t i = 0; uintmax_t j = 0;
  if (n == 0)  return ztdd_bot();
  my_val v = x[0][depth];
  if (v == INFTY) return ztdd_top();
  if(x[n-1][depth] == v) {
    if (v < 0) {i = n; j = n;}
    else if (v > 0) {i = 0; j = n;}
    else assert(0);
  } else {
    if(v > 0) {
      i = 0;
      for(j = 1; j < n && x[j][depth] == v; j *= 2) ;
      if(j < n) j = my_bsearch(x, j/2,   j, v, depth);
      else      j = my_bsearch(x, j/2, n-1, v, depth);
    } else {
      for(i = 1; i < n && x[i][depth] == v; i *= 2) ;
      if(i < n) i = my_bsearch(x, i/2,   i, v, depth);
      else      i = my_bsearch(x, i/2, n-1, v, depth);

      uintmax_t w = -v; 
      if(x[i][depth] != w) {j = i;} 
      else if (x[n-1][depth] == w) {j = n;} 
      else {
	for(j = i; j < n && x[j][depth] == w; j *= 2) ;
	if(j < n) j = my_bsearch(x, my_max(i,j/2),   j, w, depth);
	else      j = my_bsearch(x, my_max(i,j/2), n-1, w, depth);
      }
    }
  }

  my_ztddp r = ztdd_getnode(abs(v), 
			   tpconstruct(x+j, n-j, depth),
			   tpconstruct(x, i, depth+1),
			   tpconstruct(x+i, j-i, depth+1));
  ENSURE_TRUE_MSG(r != ZTDD_NULL, "Z-TDD operation failed");
  return r;
}


/**  \brief  export Z-TDD as a DIMACS cnf format.
 *  \param maxval  the maximum value 
 *  \param  f     a root node 
 *  \param  out   pointer to a dafa file
 *  \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 */
int ztdd_export_dimacs(my_val maxval, my_ztddp f, FILE *out, char *name)
{
  assert(out != NULL);
  
  if(f == ztdd_top()) {
    fprintf(out,"\n");
    return MY_SUCCESS;
  }
  if(f == ZTDD_NULL || f == ztdd_bot()) return MY_SUCCESS;

  fseek(out, 0L, SEEK_SET);

  fprintf(out, "c FILE: %s\n", name); // Insert comment line
  time_t timer;
  timer = time(NULL);
  fprintf(out, "c %s", asctime(localtime(&timer)));
  fprintf(out, "p cnf %jd %ju\n", maxval, ztdd_card(f)); // Insert problem line

  my_ztddp *br = (my_ztddp*)malloc(sizeof(my_ztddp)*(maxval+1)*2);
  ENSURE_TRUE_MSG(br != NULL, "memory allocation failed");
  my_ztddp *bp = br;
  char *sr = (char*)malloc(sizeof(char)*(maxval+1)*2);
  ENSURE_TRUE_MSG(sr != NULL, "memory allocation failed");
  char *sp = sr;
  my_val *hi = (my_val*)malloc(sizeof(my_val)*(maxval+1));
  ENSURE_TRUE_MSG(hi != NULL, "memory allocation failed");
  my_val *hp = hi;

  while(1) {
    while(f != ztdd_top()) {
      my_ztddp no = ztdd_ZERO(f);
      my_ztddp ne = ztdd_NEG(f);
      my_ztddp po = ztdd_POS(f);
      my_val fv = ztdd_label(f);
      if(ne == ztdd_bot()) {
	assert(po != ztdd_bot());
	fprintf(out, "%jd ", fv);
	assert(hp < hi + maxval+1);
	*hp = fv; hp++;
	if(no != ztdd_bot()) {assert(bp < br +(maxval+1)*2); *bp = f; bp++; *sp = 0; sp++;}
	f = po;	
      } else {
	fprintf(out, "-%jd ", fv);
	assert(hp < hi + maxval+1);
	*hp = -1*fv; hp++;
	if(po != ztdd_bot()) {assert(bp < br +(maxval+1)*2); *bp = f; bp++; *sp = 1; sp++;}
	if(no != ztdd_bot()) {assert(bp < br +(maxval+1)*2); *bp = f; bp++; *sp = 0; sp++;}
	f = ne;
      }
    }
    fprintf(out, "0 \n");
    if(bp <= br) break;
    bp--; sp--;
    my_val bv = ztdd_label(*bp);
    for(hp = hi; abs(*hp) < bv; hp++) fprintf(out, "%jd ", *hp);
    if(*sp == 1) {
      f = ztdd_POS(*bp);
      fprintf(out, "%jd ", bv);
      assert(hp < hi + maxval+1);
      *hp = bv; hp++;   
    }
    else if (*sp == 0) {f = ztdd_ZERO(*bp);}
    else assert(0);
  }

  free(br); free(sr); free(hi);
  return MY_SUCCESS;
}

